from unittest.mock import MagicMock, patch

from django.http import HttpResponse

from axes.decorators import axes_dispatch, axes_form_invalid
from tests.base import AxesTestCase


class DecoratorTestCase(AxesTestCase):
    SUCCESS_RESPONSE = HttpResponse(status=200, content="Dispatched")
    LOCKOUT_RESPONSE = HttpResponse(status=429, content="Locked out")

    def setUp(self):
        self.request = MagicMock()
        self.cls = MagicMock(return_value=self.request)
        self.func = MagicMock(return_value=self.SUCCESS_RESPONSE)

    @patch("axes.handlers.proxy.AxesProxyHandler.is_allowed", return_value=False)
    @patch("axes.decorators.get_lockout_response", return_value=LOCKOUT_RESPONSE)
    def test_axes_dispatch_locks_out(self, _, __):
        response = axes_dispatch(self.func)(self.request)
        self.assertEqual(response.content, self.LOCKOUT_RESPONSE.content)

    @patch("axes.handlers.proxy.AxesProxyHandler.is_allowed", return_value=True)
    @patch("axes.decorators.get_lockout_response", return_value=LOCKOUT_RESPONSE)
    def test_axes_dispatch_dispatches(self, _, __):
        response = axes_dispatch(self.func)(self.request)
        self.assertEqual(response.content, self.SUCCESS_RESPONSE.content)

    @patch("axes.handlers.proxy.AxesProxyHandler.is_allowed", return_value=False)
    @patch("axes.decorators.get_lockout_response", return_value=LOCKOUT_RESPONSE)
    def test_axes_form_invalid_locks_out(self, _, __):
        response = axes_form_invalid(self.func)(self.cls)
        self.assertEqual(response.content, self.LOCKOUT_RESPONSE.content)

    @patch("axes.handlers.proxy.AxesProxyHandler.is_allowed", return_value=True)
    @patch("axes.decorators.get_lockout_response", return_value=LOCKOUT_RESPONSE)
    def test_axes_form_invalid_dispatches(self, _, __):
        response = axes_form_invalid(self.func)(self.cls)
        self.assertEqual(response.content, self.SUCCESS_RESPONSE.content)



# Generated by Django 3.2.12 on 2022-03-15 03:00

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("axes", "0007_alter_accessattempt_unique_together"),
    ]

    operations = [
        migrations.CreateModel(
            name="AccessFailureLog",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "user_agent",
                    models.CharField(
                        db_index=True, max_length=255, verbose_name="User Agent"
                    ),
                ),
                (
                    "ip_address",
                    models.GenericIPAddressField(
                        db_index=True, null=True, verbose_name="IP Address"
                    ),
                ),
                (
                    "username",
                    models.CharField(
                        db_index=True,
                        max_length=255,
                        null=True,
                        verbose_name="Username",
                    ),
                ),
                (
                    "http_accept",
                    models.CharField(max_length=1025, verbose_name="HTTP Accept"),
                ),
                ("path_info", models.CharField(max_length=255, verbose_name="Path")),
                (
                    "attempt_time",
                    models.DateTimeField(
                        auto_now_add=True, verbose_name="Attempt Time"
                    ),
                ),
                (
                    "locked_out",
                    models.BooleanField(
                        blank=True, default=False, verbose_name="Access lock out"
                    ),
                ),
            ],
            options={
                "verbose_name": "access failure",
                "verbose_name_plural": "access failures",
            },
        ),
    ]



from django.core.management.base import BaseCommand

from axes.utils import reset


class Command(BaseCommand):
    help = "Reset all access attempts and lockouts for a given IP address and username"

    def add_arguments(self, parser):
        parser.add_argument("ip", type=str)
        parser.add_argument("username", type=str)

    def handle(self, *args, **options):
        count = reset(ip=options["ip"], username=options["username"])

        if count:
            self.stdout.write(f"{count} attempts removed.")
        else:
            self.stdout.write("No attempts found.")



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = []

    operations = [
        migrations.CreateModel(
            name="AccessAttempt",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("user_agent", models.CharField(max_length=255)),
                (
                    "ip_address",
                    models.GenericIPAddressField(null=True, verbose_name="IP Address"),
                ),
                ("username", models.CharField(max_length=255, null=True)),
                ("trusted", models.BooleanField(default=False)),
                (
                    "http_accept",
                    models.CharField(max_length=1025, verbose_name="HTTP Accept"),
                ),
                ("path_info", models.CharField(max_length=255, verbose_name="Path")),
                ("attempt_time", models.DateTimeField(auto_now_add=True)),
                ("get_data", models.TextField(verbose_name="GET Data")),
                ("post_data", models.TextField(verbose_name="POST Data")),
                (
                    "failures_since_start",
                    models.PositiveIntegerField(verbose_name="Failed Logins"),
                ),
            ],
            options={"ordering": ["-attempt_time"], "abstract": False},
        ),
        migrations.CreateModel(
            name="AccessLog",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("user_agent", models.CharField(max_length=255)),
                (
                    "ip_address",
                    models.GenericIPAddressField(null=True, verbose_name="IP Address"),
                ),
                ("username", models.CharField(max_length=255, null=True)),
                ("trusted", models.BooleanField(default=False)),
                (
                    "http_accept",
                    models.CharField(max_length=1025, verbose_name="HTTP Accept"),
                ),
                ("path_info", models.CharField(max_length=255, verbose_name="Path")),
                ("attempt_time", models.DateTimeField(auto_now_add=True)),
                ("logout_time", models.DateTimeField(null=True, blank=True)),
            ],
            options={"ordering": ["-attempt_time"], "abstract": False},
        ),
    ]



"""
Axes utility functions that are publicly available.

This module is separate for historical reasons
and offers a backwards compatible import path.
"""

from logging import getLogger
from typing import Optional

from django.http import HttpRequest

from axes.handlers.proxy import AxesProxyHandler
from axes.helpers import get_client_ip_address, get_lockout_parameters

log = getLogger(__name__)


def reset(
    ip: Optional[str] = None, username: Optional[str] = None, ip_or_username=False
) -> int:
    """
    Reset records that match IP or username, and return the count of removed attempts.

    This utility method is meant to be used from the CLI or via Python API.
    """

    return AxesProxyHandler.reset_attempts(
        ip_address=ip, username=username, ip_or_username=ip_or_username
    )


def reset_request(request: HttpRequest) -> int:
    """
    Reset records that match IP or username, and return the count of removed attempts.

    This utility method is meant to be used from the CLI or via Python API.
    """
    lockout_paramaters = get_lockout_parameters(request)

    ip: Optional[str] = get_client_ip_address(request)
    username = request.GET.get("username", None)

    ip_required = False
    username_required = False
    ip_and_username = False

    for param in lockout_paramaters:
        # hack: in works with all iterables, including strings
        # so this checks works with separate parameters
        # and with parameters combinations
        if "username" in param and "ip_address" in param:
            ip_and_username = True
            ip_required = True
            username_required = True
            break
        if "username" in param:
            username_required = True
        elif "ip_address" in param:
            ip_required = True

    ip_or_username = not ip_and_username and ip_required and username_required
    if not ip_required:
        ip = None
    if not username_required:
        username = None

    if not ip and not username:
        return 0
        # We don't want to reset everything, if there is some wrong request parameter

    # TODO: reset based on user_agent?
    return reset(ip, username, ip_or_username)



from django.db import models, migrations


class Migration(migrations.Migration):
    dependencies = [("axes", "0002_auto_20151217_2044")]

    operations = [
        migrations.AlterField(
            model_name="accessattempt",
            name="failures_since_start",
            field=models.PositiveIntegerField(verbose_name="Failed Logins"),
        ),
        migrations.AlterField(
            model_name="accessattempt",
            name="get_data",
            field=models.TextField(verbose_name="GET Data"),
        ),
        migrations.AlterField(
            model_name="accessattempt",
            name="http_accept",
            field=models.CharField(verbose_name="HTTP Accept", max_length=1025),
        ),
        migrations.AlterField(
            model_name="accessattempt",
            name="ip_address",
            field=models.GenericIPAddressField(
                null=True, verbose_name="IP Address", db_index=True
            ),
        ),
        migrations.AlterField(
            model_name="accessattempt",
            name="path_info",
            field=models.CharField(verbose_name="Path", max_length=255),
        ),
        migrations.AlterField(
            model_name="accessattempt",
            name="post_data",
            field=models.TextField(verbose_name="POST Data"),
        ),
        migrations.AlterField(
            model_name="accesslog",
            name="http_accept",
            field=models.CharField(verbose_name="HTTP Accept", max_length=1025),
        ),
        migrations.AlterField(
            model_name="accesslog",
            name="ip_address",
            field=models.GenericIPAddressField(
                null=True, verbose_name="IP Address", db_index=True
            ),
        ),
        migrations.AlterField(
            model_name="accesslog",
            name="path_info",
            field=models.CharField(verbose_name="Path", max_length=255),
        ),
    ]



# Generated by Django 4.2.2 on 2024-04-30 07:57

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("axes", "0008_accessfailurelog"),
    ]

    operations = [
        migrations.AddField(
            model_name="accesslog",
            name="session_hash",
            field=models.CharField(
                blank=True,
                default="",
                max_length=64,
                verbose_name="Session key hash (sha256)",
            ),
        ),
    ]



"""
Integration tests for the login handling.

TODO: Clean up the tests in this module.
"""
from datetime import timedelta
from importlib import import_module
from time import sleep

from django.contrib.auth import get_user_model, login, logout
from django.http import HttpRequest
from django.test import override_settings, TestCase
from django.urls import reverse

from axes.conf import settings
from axes.helpers import get_cache, make_cache_key_list, get_cool_off, get_failure_limit
from axes.models import AccessAttempt
from tests.base import AxesTestCase


class DjangoLoginTestCase(TestCase):
    def setUp(self):
        engine = import_module(settings.SESSION_ENGINE)

        self.request = HttpRequest()
        self.request.session = engine.SessionStore()

        self.username = "john.doe"
        self.password = "hunter2"

        self.user = get_user_model().objects.create(username=self.username, is_staff=True)
        self.user.set_password(self.password)
        self.user.save()
        self.user.backend = "django.contrib.auth.backends.ModelBackend"


class DjangoContribAuthLoginTestCase(DjangoLoginTestCase):
    def test_login(self):
        login(self.request, self.user)

    def test_logout(self):
        login(self.request, self.user)
        logout(self.request)


@override_settings(AXES_ENABLED=False)
class DjangoTestClientLoginTestCase(DjangoLoginTestCase):
    def test_client_login(self):
        self.client.login(username=self.username, password=self.password)
        response = self.client.get(reverse("admin:index"))
        self.assertEqual(response.status_code, 200)

    def test_client_logout(self):
        self.client.login(username=self.username, password=self.password)
        self.client.logout()
        response = self.client.get(reverse("admin:index"))
        self.assertEqual(response.status_code, 302)

    def test_client_force_login(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse("admin:index"))
        self.assertEqual(response.status_code, 200)


class DatabaseLoginTestCase(AxesTestCase):
    """
    Test for lockouts under different configurations and circumstances to prevent false positives and false negatives.

    Always block attempted logins for the same user from the same IP.
    Always allow attempted logins for a different user from a different IP.
    """

    IP_1 = "10.1.1.1"
    IP_2 = "10.2.2.2"
    IP_3 = "10.2.2.3"
    USER_1 = "valid-user-1"
    USER_2 = "valid-user-2"
    USER_3 = "valid-user-3"
    EMAIL_1 = "valid-email-1@example.com"
    EMAIL_2 = "valid-email-2@example.com"

    VALID_USERNAME = USER_1
    VALID_EMAIL = EMAIL_1
    VALID_PASSWORD = "valid-password"

    VALID_IP_ADDRESS = IP_1

    WRONG_PASSWORD = "wrong-password"
    LOCKED_MESSAGE = "Account locked: too many login attempts."
    LOGIN_FORM_KEY = '<input type="submit" value="Log in" />'
    ATTEMPT_NOT_BLOCKED = 200
    ALLOWED = 302
    BLOCKED = 429

    def _login(self, username, password, ip_addr="127.0.0.1", user_agent="test-browser", **kwargs):
        """
        Login a user and get the response.

        IP address can be configured to test IP blocking functionality.
        """

        post_data = {"username": username, "password": password}

        post_data.update(kwargs)

        return self.client.post(
            reverse("admin:login"),
            post_data,
            REMOTE_ADDR=ip_addr,
            HTTP_USER_AGENT=user_agent,
        )

    def _lockout_user_from_ip(self, username, ip_addr, user_agent="test-browser"):
        for _ in range(settings.AXES_FAILURE_LIMIT):
            response = self._login(
                username=username, password=self.WRONG_PASSWORD, ip_addr=ip_addr, user_agent=user_agent,
            )
        return response

    def _lockout_user1_from_ip1(self):
        return self._lockout_user_from_ip(username=self.USER_1, ip_addr=self.IP_1)

    def setUp(self):
        """
        Create two valid users for authentication.
        """

        super().setUp()

        self.user2 = get_user_model().objects.create_superuser(
            username=self.USER_2,
            email=self.EMAIL_2,
            password=self.VALID_PASSWORD,
            is_staff=True,
            is_superuser=True,
        )

    def test_login(self):
        """
        Test a valid login for a real username.
        """

        response = self._login(self.username, self.password)
        self.assertNotContains(
            response, self.LOGIN_FORM_KEY, status_code=self.ALLOWED, html=True
        )

    def test_lockout_limit_once(self):
        """
        Test the login lock trying to login one more time than failure limit.
        """

        response = self.lockout()
        self.assertContains(response, self.LOCKED_MESSAGE, status_code=self.BLOCKED)

    def test_lockout_limit_many(self):
        """
        Test the login lock trying to login a lot of times more than failure limit.
        """

        self.lockout()

        for _ in range(settings.AXES_FAILURE_LIMIT):
            response = self.login()
            self.assertContains(response, self.LOCKED_MESSAGE, status_code=self.BLOCKED)

    def attempt_count(self):
        return AccessAttempt.objects.count()

    @override_settings(AXES_RESET_ON_SUCCESS=False)
    def test_reset_on_success_false(self):
        self.almost_lockout()
        self.login(is_valid_username=True, is_valid_password=True)

        response = self.login()
        self.assertContains(response, self.LOCKED_MESSAGE, status_code=self.BLOCKED)
        self.assertTrue(self.attempt_count())

    @override_settings(AXES_RESET_ON_SUCCESS=True)
    def test_reset_on_success_true(self):
        self.almost_lockout()
        self.assertTrue(self.attempt_count())

        self.login(is_valid_username=True, is_valid_password=True)
        self.assertFalse(self.attempt_count())

        response = self.lockout()
        self.assertContains(response, self.LOCKED_MESSAGE, status_code=self.BLOCKED)
        self.assertTrue(self.attempt_count())

    @override_settings(AXES_LOCKOUT_PARAMETERS=[["username", "ip_address"]])
    def test_lockout_by_combination_user_and_ip(self):
        """
        Test login failure when lockout parameters is combination
        of username and ip_address.
        """

        # test until one try before the limit
        for _ in range(1, settings.AXES_FAILURE_LIMIT):
            response = self.login(is_valid_username=True, is_valid_password=False)
            # Check if we are in the same login page
            self.assertContains(response, self.LOGIN_FORM_KEY, html=True)

        # So, we shouldn't have gotten a lock-out yet.
        # But we should get one now
        response = self.login(is_valid_username=True, is_valid_password=False)
        self.assertContains(response, self.LOCKED_MESSAGE, status_code=429)

    @override_settings(AXES_LOCKOUT_PARAMETERS=["username"])
    def test_lockout_by_only_user_failures(self):
        """
        Test login failure when lockout parameter is username.
        """

        # test until one try before the limit
        for _ in range(1, settings.AXES_FAILURE_LIMIT):
            response = self._login(self.username, self.WRONG_PASSWORD)

            # Check if we are in the same login page
            self.assertContains(response, self.LOGIN_FORM_KEY, html=True)

        # So, we shouldn't have gotten a lock-out yet.
        # But we should get one now
        response = self._login(self.username, self.WRONG_PASSWORD)
        self.assertContains(response, self.LOCKED_MESSAGE, status_code=self.BLOCKED)

        # reset the username only and make sure we can log in now even though our IP has failed each time
        self.reset(username=self.username)

        response = self._login(self.username, self.password)

        # Check if we are still in the login page
        self.assertNotContains(
            response, self.LOGIN_FORM_KEY, status_code=self.ALLOWED, html=True
        )

        # now create failure_limit + 1 failed logins and then we should still
        # be able to login with valid_username
        for _ in range(settings.AXES_FAILURE_LIMIT):
            response = self._login(self.username, self.password)

        # Check if we can still log in with valid user
        response = self._login(self.username, self.password)
        self.assertNotContains(
            response, self.LOGIN_FORM_KEY, status_code=self.ALLOWED, html=True
        )

    @override_settings(AXES_LOCKOUT_PARAMETERS=["user_agent"])
    def test_lockout_by_user_agent_only(self):
        """
        Test login failure when lockout parameter is only user_agent
        """
        # User is locked out with "test-browser" user agent.
        self._lockout_user_from_ip(username="username", ip_addr=self.IP_1, user_agent="test-browser")

        # Test he is locked:
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_1, user_agent="test-browser")
        self.assertEqual(response.status_code, self.BLOCKED)

        # Test he is locked with another username:
        response = self._login("username2", self.VALID_PASSWORD, ip_addr=self.IP_1, user_agent="test-browser")
        self.assertEqual(response.status_code, self.BLOCKED)

        # Test he is locked with another ip:
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_2, user_agent="test-browser")
        self.assertEqual(response.status_code, self.BLOCKED)

        # Test with another user agent:
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_1, user_agent="test-browser-2")
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=["ip_address", "username", "user_agent"])
    def test_lockout_by_all_parameters(self):
        # User is locked out with "test-browser" user agent.
        self._lockout_user_from_ip(username="username", ip_addr=self.IP_1, user_agent="test-browser")

        # Test he is locked:
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_1, user_agent="test-browser")
        self.assertEqual(response.status_code, self.BLOCKED)

        # Test he is locked by username:
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_2, user_agent="test-browser2")
        self.assertEqual(response.status_code, self.BLOCKED)

        # Test he is locked by ip:
        response = self._login("username2", self.VALID_PASSWORD, ip_addr=self.IP_1, user_agent="test-browser2")
        self.assertEqual(response.status_code, self.BLOCKED)

        # Test he is locked by user_agent:
        response = self._login("username2", self.VALID_PASSWORD, ip_addr=self.IP_2, user_agent="test-browser")
        self.assertEqual(response.status_code, self.BLOCKED)

        # Test he is allowed to login with different username, ip and user_agent
        response = self._login("username2", self.VALID_PASSWORD, ip_addr=self.IP_2, user_agent="test-browser2")
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=[["ip_address", "username", "user_agent"]])
    def test_lockout_by_combination_of_all_parameters(self):
        # User is locked out with "test-browser" user agent.
        self._lockout_user_from_ip(username="username", ip_addr=self.IP_1, user_agent="test-browser")

        # Test he is locked:
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_1, user_agent="test-browser")
        self.assertEqual(response.status_code, self.BLOCKED)

        # Test he is allowed to login with different username:
        response = self._login("username2", self.VALID_PASSWORD, ip_addr=self.IP_1, user_agent="test-browser")
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # Test he is allowed to login with different IP:
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_2, user_agent="test-browser")
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # Test he is allowed to login with different user_agent:
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_1, user_agent="test-browser2")
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # Test he is allowed to login with different username, ip and user_agent
        response = self._login("username2", self.VALID_PASSWORD, ip_addr=self.IP_2, user_agent="test-browser2")
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=["ip_address", ["username", "user_agent"]])
    def test_lockout_by_ip_or_username_and_user_agent(self):
        # User is locked out with "test-browser" user agent.
        self._lockout_user_from_ip(username="username", ip_addr=self.IP_1, user_agent="test-browser")

        # Test he is locked:
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_1, user_agent="test-browser")
        self.assertEqual(response.status_code, self.BLOCKED)

        # Test he is locked by ip:
        response = self._login("username2", self.VALID_PASSWORD, ip_addr=self.IP_1, user_agent="test-browser2")
        self.assertEqual(response.status_code, self.BLOCKED)

        # Test he is locked by username and user_agent:
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_2, user_agent="test-browser")
        self.assertEqual(response.status_code, self.BLOCKED)

        # Test he is allowed to login with different username and ip
        response = self._login("username2", self.VALID_PASSWORD, ip_addr=self.IP_2, user_agent="test-browser")
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # Test he is allowed to login with different user_agent and ip
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_2, user_agent="test-browser2")
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # Test he is allowed to login with different username, ip and user_agent
        response = self._login("username2", self.VALID_PASSWORD, ip_addr=self.IP_2, user_agent="test-browser2")
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=[["ip_address", "user_agent"], ["username", "user_agent"]])
    def test_lockout_by_ip_and_user_agent_or_username_and_user_agent(self):
        # User is locked out with "test-browser" user agent.
        self._lockout_user_from_ip(username="username", ip_addr=self.IP_1, user_agent="test-browser")

        # Test he is locked:
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_1, user_agent="test-browser")
        self.assertEqual(response.status_code, self.BLOCKED)

        # Test he is locked by ip and user_agent:
        response = self._login("username2", self.VALID_PASSWORD, ip_addr=self.IP_1, user_agent="test-browser")
        self.assertEqual(response.status_code, self.BLOCKED)

        # Test he is locked by username and user_agent:
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_2, user_agent="test-browser")
        self.assertEqual(response.status_code, self.BLOCKED)

        # Test he is allowed to login with different username and ip
        response = self._login("username2", self.VALID_PASSWORD, ip_addr=self.IP_2, user_agent="test-browser")
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # Test he is allowed to login with different user_agent
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_1, user_agent="test-browser2")
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # Test he is allowed to login with different username, ip and user_agent
        response = self._login("username2", self.VALID_PASSWORD, ip_addr=self.IP_2, user_agent="test-browser2")
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)


    # Test for true and false positives when blocking by IP *OR* user (default)
    # Cache disabled. Default settings.
    def test_lockout_by_ip_blocks_when_same_user_same_ip_without_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 1 is still blocked from IP 1.
        response = self._login(self.USER_1, self.VALID_PASSWORD, ip_addr=self.IP_1)
        self.assertEqual(response.status_code, self.BLOCKED)

    def test_lockout_by_ip_allows_when_same_user_diff_ip_without_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 1 can still login from IP 2.
        response = self._login(self.USER_1, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.ALLOWED)

    def test_lockout_by_ip_blocks_when_diff_user_same_ip_without_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 2 is also locked out from IP 1.
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_1)
        self.assertEqual(response.status_code, self.BLOCKED)

    def test_lockout_by_ip_allows_when_diff_user_diff_ip_without_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 2 can still login from IP 2.
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.ALLOWED)

    # Test for true and false positives when blocking by user only.
    # Cache disabled. When AXES_ONLY_USER_FAILURES = True
    @override_settings(AXES_LOCKOUT_PARAMETERS=["username"])
    def test_lockout_by_user_blocks_when_same_user_same_ip_without_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 1 is still blocked from IP 1.
        response = self._login(self.USER_1, self.VALID_PASSWORD, ip_addr=self.IP_1)
        self.assertEqual(response.status_code, self.BLOCKED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=["username"])
    def test_lockout_by_user_blocks_when_same_user_diff_ip_without_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 1 is also locked out from IP 2.
        response = self._login(self.USER_1, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.BLOCKED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=["username"])
    def test_lockout_by_user_allows_when_diff_user_same_ip_without_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 2 can still login from IP 1.
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_1)
        self.assertEqual(response.status_code, self.ALLOWED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=["username"])
    def test_lockout_by_user_allows_when_diff_user_diff_ip_without_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 2 can still login from IP 2.
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.ALLOWED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=["username"])
    def test_lockout_by_user_with_empty_username_allows_other_users_without_cache(self):
        # User with empty username is locked out from IP 1.
        self._lockout_user_from_ip(username="", ip_addr=self.IP_1)

        # Still possible to access the login page
        response = self.client.get(reverse("admin:login"), REMOTE_ADDR=self.IP_1)
        self.assertContains(response, self.LOGIN_FORM_KEY, status_code=200, html=True)

    # Test for true and false positives when blocking by user and IP together.
    # Cache disabled. When LOCK_OUT_BY_COMBINATION_USER_AND_IP = True
    @override_settings(AXES_LOCKOUT_PARAMETERS=[["username", "ip_address"]])
    def test_lockout_by_user_and_ip_blocks_when_same_user_same_ip_without_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 1 is still blocked from IP 1.
        response = self._login(self.USER_1, self.VALID_PASSWORD, ip_addr=self.IP_1)
        self.assertEqual(response.status_code, self.BLOCKED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=[["username", "ip_address"]])
    def test_lockout_by_user_and_ip_allows_when_same_user_diff_ip_without_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 1 can still login from IP 2.
        response = self._login(self.USER_1, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.ALLOWED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=[["username", "ip_address"]])
    def test_lockout_by_user_and_ip_allows_when_diff_user_same_ip_without_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 2 can still login from IP 1.
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_1)
        self.assertEqual(response.status_code, self.ALLOWED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=[["username", "ip_address"]])
    def test_lockout_by_user_and_ip_allows_when_diff_user_diff_ip_without_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 2 can still login from IP 2.
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.ALLOWED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=[["username", "ip_address"]])
    def test_lockout_by_user_and_ip_with_empty_username_allows_other_users_without_cache(
        self,
    ):
        # User with empty username is locked out from IP 1.
        self._lockout_user_from_ip(username="", ip_addr=self.IP_1)

        # Still possible to access the login page
        response = self.client.get(reverse("admin:login"), REMOTE_ADDR=self.IP_1)
        self.assertContains(response, self.LOGIN_FORM_KEY, status_code=200, html=True)

    @override_settings(AXES_LOCKOUT_PARAMETERS=[["ip_address", "user_agent"]])
    def test_lockout_by_user_still_allows_login_with_differnet_user_agent(self):
        # User with empty username is locked out with "test-browser" user agent.
        self._lockout_user_from_ip(username="username", ip_addr=self.IP_1, user_agent="test-browser")

        # Test he is locked:
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_1, user_agent="test-browser")
        self.assertEqual(response.status_code, self.BLOCKED)

        # Test with another user agent:
        response = self._login("username", self.VALID_PASSWORD, ip_addr=self.IP_1, user_agent="test-browser-2")
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

    # Test for true and false positives when blocking by IP *OR* user (default)
    # With cache enabled. Default criteria.
    def test_lockout_by_ip_blocks_when_same_user_same_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 1 is still blocked from IP 1.
        response = self._login(self.USER_1, self.VALID_PASSWORD, ip_addr=self.IP_1)
        self.assertEqual(response.status_code, self.BLOCKED)

    def test_lockout_by_ip_allows_when_same_user_diff_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 1 can still login from IP 2.
        response = self._login(self.USER_1, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.ALLOWED)

    def test_lockout_by_ip_blocks_when_diff_user_same_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 2 is also locked out from IP 1.
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_1)
        self.assertEqual(response.status_code, self.BLOCKED)

    def test_lockout_by_ip_allows_when_diff_user_diff_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 2 can still login from IP 2.
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.ALLOWED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=["username"])
    def test_lockout_by_user_with_empty_username_allows_other_users_using_cache(self):
        # User with empty username is locked out from IP 1.
        self._lockout_user_from_ip(username="", ip_addr=self.IP_1)

        # Still possible to access the login page
        response = self.client.get(reverse("admin:login"), REMOTE_ADDR=self.IP_1)
        self.assertContains(response, self.LOGIN_FORM_KEY, status_code=200, html=True)

    # Test for true and false positives when blocking by user only.
    # With cache enabled. When AXES_ONLY_USER_FAILURES = True
    @override_settings(AXES_LOCKOUT_PARAMETERS=["username"])
    def test_lockout_by_user_blocks_when_same_user_same_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 1 is still blocked from IP 1.
        response = self._login(self.USER_1, self.VALID_PASSWORD, ip_addr=self.IP_1)
        self.assertEqual(response.status_code, self.BLOCKED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=["username"])
    def test_lockout_by_user_blocks_when_same_user_diff_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 1 is also locked out from IP 2.
        response = self._login(self.USER_1, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.BLOCKED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=["username"])
    def test_lockout_by_user_allows_when_diff_user_same_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 2 can still login from IP 1.
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_1)
        self.assertEqual(response.status_code, self.ALLOWED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=["username"])
    def test_lockout_by_user_allows_when_diff_user_diff_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 2 can still login from IP 2.
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.ALLOWED)

    # Test for true and false positives when blocking by user and IP together.
    # With cache enabled. When LOCK_OUT_BY_COMBINATION_USER_AND_IP = True
    @override_settings(AXES_LOCKOUT_PARAMETERS=[["username", "ip_address"]])
    def test_lockout_by_user_and_ip_blocks_when_same_user_same_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 1 is still blocked from IP 1.
        response = self._login(self.USER_1, self.VALID_PASSWORD, ip_addr=self.IP_1)
        self.assertEqual(response.status_code, self.BLOCKED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=[["username", "ip_address"]])
    def test_lockout_by_user_and_ip_allows_when_same_user_diff_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 1 can still login from IP 2.
        response = self._login(self.USER_1, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.ALLOWED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=[["username", "ip_address"]])
    def test_lockout_by_user_and_ip_allows_when_diff_user_same_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 2 can still login from IP 1.
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_1)
        self.assertEqual(response.status_code, self.ALLOWED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=[["username", "ip_address"]])
    def test_lockout_by_user_and_ip_allows_when_diff_user_diff_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 2 can still login from IP 2.
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.ALLOWED)

    @override_settings(
        AXES_LOCKOUT_PARAMETERS=[["username", "ip_address"]], AXES_FAILURE_LIMIT=2
    )
    def test_lockout_by_user_and_ip_allows_when_diff_user_same_ip_using_cache_multiple_attempts(
        self,
    ):
        # User 1 is locked out from IP 1.
        response = self._login(self.USER_1, self.WRONG_PASSWORD, self.IP_1)
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # Second attempt from different IP
        response = self._login(self.USER_1, self.WRONG_PASSWORD, self.IP_2)
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # Second attempt from same IP, different username
        response = self._login(self.USER_2, self.WRONG_PASSWORD, self.IP_1)
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # User 1 is blocked from IP 1
        response = self._login(self.USER_1, self.WRONG_PASSWORD, ip_addr=self.IP_1)
        self.assertContains(response, self.LOCKED_MESSAGE, status_code=self.BLOCKED)

        # User 1 is blocked from IP 2
        response = self._login(self.USER_1, self.WRONG_PASSWORD, ip_addr=self.IP_2)
        self.assertContains(response, self.LOCKED_MESSAGE, status_code=self.BLOCKED)

        # User 2 can still login from IP 2, only he has 1 attempt left
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.ALLOWED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=[["username", "ip_address"]])
    def test_lockout_by_user_and_ip_with_empty_username_allows_other_users_using_cache(
        self,
    ):
        # User with empty username is locked out from IP 1.
        self._lockout_user_from_ip(username="", ip_addr=self.IP_1)

        # Still possible to access the login page
        response = self.client.get(reverse("admin:login"), REMOTE_ADDR=self.IP_1)
        self.assertContains(response, self.LOGIN_FORM_KEY, status_code=200, html=True)

    # Test for true and false positives when blocking by user or IP together.
    # With cache enabled. When AXES_LOCK_OUT_BY_USER_OR_IP = True
    @override_settings(AXES_LOCKOUT_PARAMETERS=["username", "ip_address"])
    def test_lockout_by_user_or_ip_blocks_when_same_user_same_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 1 is still blocked from IP 1.
        response = self._login(self.USER_1, self.VALID_PASSWORD, ip_addr=self.IP_1)
        self.assertEqual(response.status_code, self.BLOCKED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=["username", "ip_address"])
    def test_lockout_by_user_or_ip_allows_when_same_user_diff_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 1 is blocked out from IP 1
        response = self._login(self.USER_1, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.BLOCKED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=["username", "ip_address"])
    def test_lockout_by_user_or_ip_allows_when_diff_user_same_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 2 can still login from IP 1.
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_1)
        self.assertEqual(response.status_code, self.BLOCKED)

    @override_settings(
        AXES_LOCKOUT_PARAMETERS=["username", "ip_address"], AXES_FAILURE_LIMIT=3
    )
    def test_lockout_by_user_or_ip_allows_when_diff_user_same_ip_using_cache_multiple_attempts(
        self,
    ):
        # User 1 is locked out from IP 1.
        response = self._login(self.USER_1, self.WRONG_PASSWORD, self.IP_1)
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # Second attempt from different IP
        response = self._login(self.USER_1, self.WRONG_PASSWORD, self.IP_2)
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # User 1 is blocked on all IPs, he reached 2 attempts
        response = self._login(self.USER_1, self.WRONG_PASSWORD, ip_addr=self.IP_2)
        self.assertContains(response, self.LOCKED_MESSAGE, status_code=self.BLOCKED)
        response = self._login(self.USER_1, self.WRONG_PASSWORD, ip_addr=self.IP_3)
        self.assertContains(response, self.LOCKED_MESSAGE, status_code=self.BLOCKED)

        # IP 1 has still one attempt left
        response = self._login(self.USER_2, self.WRONG_PASSWORD, self.IP_1)
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # But now IP 1 is blocked for all attempts
        response = self._login(self.USER_1, self.WRONG_PASSWORD, ip_addr=self.IP_1)
        self.assertContains(response, self.LOCKED_MESSAGE, status_code=self.BLOCKED)
        response = self._login(self.USER_2, self.WRONG_PASSWORD, ip_addr=self.IP_1)
        self.assertContains(response, self.LOCKED_MESSAGE, status_code=self.BLOCKED)
        response = self._login(self.USER_3, self.WRONG_PASSWORD, ip_addr=self.IP_1)
        self.assertContains(response, self.LOCKED_MESSAGE, status_code=self.BLOCKED)

    @override_settings(
        AXES_LOCKOUT_PARAMETERS=["username", "ip_address"], AXES_FAILURE_LIMIT=3
    )
    def test_lockout_by_user_or_ip_allows_when_diff_user_same_ip_using_cache_multiple_failed_attempts(
        self,
    ):
        """Test, if the failed attempts make also impact on the attempt count"""
        # User 1 is locked out from IP 1.
        response = self._login(self.USER_1, self.WRONG_PASSWORD, self.IP_1)
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # Second attempt from different IP
        response = self._login(self.USER_1, self.WRONG_PASSWORD, self.IP_2)
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # Second attempt from same IP, different username
        response = self._login(self.USER_2, self.WRONG_PASSWORD, self.IP_1)
        self.assertEqual(response.status_code, self.ATTEMPT_NOT_BLOCKED)

        # User 1 is blocked from IP 2
        response = self._login(self.USER_1, self.WRONG_PASSWORD, ip_addr=self.IP_2)
        self.assertContains(response, self.LOCKED_MESSAGE, status_code=self.BLOCKED)

        # On IP 2 it is only 2. attempt, for user 2 it is also 2. attempt -> allow log in
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.ALLOWED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=["username", "ip_address"])
    def test_lockout_by_user_or_ip_allows_when_diff_user_diff_ip_using_cache(self):
        # User 1 is locked out from IP 1.
        self._lockout_user1_from_ip1()

        # User 2 can still login from IP 2.
        response = self._login(self.USER_2, self.VALID_PASSWORD, ip_addr=self.IP_2)
        self.assertEqual(response.status_code, self.ALLOWED)

    @override_settings(AXES_LOCKOUT_PARAMETERS=["username", "ip_address"])
    def test_lockout_by_user_or_ip_with_empty_username_allows_other_users_using_cache(
        self,
    ):
        # User with empty username is locked out from IP 1.
        self._lockout_user_from_ip(username="", ip_addr=self.IP_1)

        # Still possible to access the login page
        response = self.client.get(reverse("admin:login"), REMOTE_ADDR=self.IP_1)
        self.assertContains(response, self.LOGIN_FORM_KEY, status_code=200, html=True)

    @override_settings(
        AXES_COOLOFF_TIME=timedelta(seconds=1),
        AXES_RESET_COOL_OFF_ON_FAILURE_DURING_LOCKOUT=False,
        AXES_FAILURE_LIMIT=2,
    )
    def test_login_during_lockout_doesnt_reset_cool_off_time(self):
        # Lockout
        for _ in range(get_failure_limit(None, None)):
            self.login(self.USER_1)

        # Attempt during lockout
        sleep_time = get_cool_off().total_seconds() / 2
        sleep(sleep_time)
        self.login(self.USER_1)
        sleep(sleep_time)

        # New attempt after initial lockout period: should work
        response = self.login(is_valid_username=True, is_valid_password=True)
        self.assertNotContains(response, self.LOCKED_MESSAGE, status_code=self.ALLOWED)

    @override_settings(
        AXES_COOLOFF_TIME=timedelta(seconds=1),
        AXES_RESET_COOL_OFF_ON_FAILURE_DURING_LOCKOUT=True,
        AXES_FAILURE_LIMIT=2,
    )
    def test_login_during_lockout_does_reset_cool_off_time(self):
        # Lockout
        for _ in range(get_failure_limit(None, None)):
            self.login(self.USER_1)

        # Attempt during lockout
        sleep_time = get_cool_off().total_seconds() / 2
        sleep(sleep_time)
        self.login(self.USER_1)
        sleep(sleep_time)

        # New attempt after initial lockout period: should not work
        response = self.login(is_valid_username=True, is_valid_password=True)
        self.assertContains(response, self.LOCKED_MESSAGE, status_code=self.BLOCKED)


# Test the same logic with cache handler
@override_settings(AXES_HANDLER="axes.handlers.cache.AxesCacheHandler")
class CacheLoginTestCase(DatabaseLoginTestCase):
    def attempt_count(self):
        cache = get_cache()
        keys = cache._cache
        return len(keys)

    def reset(self, **kwargs):
        get_cache().delete(make_cache_key_list([kwargs])[0])



from typing import Optional
from django.conf import settings
from django.contrib.auth.backends import ModelBackend
from django.http import HttpRequest

from axes.exceptions import (
    AxesBackendPermissionDenied,
    AxesBackendRequestParameterRequired,
)
from axes.handlers.proxy import AxesProxyHandler
from axes.helpers import get_credentials, get_lockout_message, toggleable


class AxesStandaloneBackend:
    """
    Authentication backend class that forbids login attempts for locked out users.

    Use this class as the first item of ``AUTHENTICATION_BACKENDS`` to
    prevent locked out users from being logged in by the Django authentication flow.

    .. note:: This backend does not log your user in. It monitors login attempts.
              It also does not run any permissions checks at all.
              Authentication is handled by the following backends that are configured in ``AUTHENTICATION_BACKENDS``.
    """

    @toggleable
    def authenticate(
        self,
        request: HttpRequest,
        username: Optional[str] = None,
        password: Optional[str] = None,
        **kwargs: dict,
    ):
        """
        Checks user lockout status and raises an exception if user is not allowed to log in.

        This method interrupts the login flow and inserts  error message directly to the
        ``response_context`` attribute that is supplied as a keyword argument.

        :keyword response_context: kwarg that will be have its ``error`` attribute updated with context.
        :raises AxesBackendRequestParameterRequired: if request parameter is not passed.
        :raises AxesBackendPermissionDenied: if user is already locked out.
        """

        if request is None:
            raise AxesBackendRequestParameterRequired(
                "AxesBackend requires a request as an argument to authenticate"
            )

        credentials = get_credentials(username=username, password=password, **kwargs)

        if AxesProxyHandler.is_allowed(request, credentials):
            return

        # Locked out, don't try to authenticate, just update response_context and return.
        # Its a bit weird to pass a context and expect a response value but its nice to get a "why" back.

        error_msg = get_lockout_message()
        response_context = kwargs.get("response_context", {})
        response_context["error"] = error_msg

        # This flag can be used later to check if it was Axes that denied the login attempt.
        if not settings.AXES_RESET_COOL_OFF_ON_FAILURE_DURING_LOCKOUT:
            request.axes_locked_out = True

        # Raise an error that stops the authentication flows at django.contrib.auth.authenticate.
        # This error stops bubbling up at the authenticate call which catches backend PermissionDenied errors.
        # After this error is caught by authenticate it emits a signal indicating user login failed,
        # which is processed by axes.signals.log_user_login_failed which logs and flags the failed request.
        # The axes.middleware.AxesMiddleware further processes the flagged request into a readable response.

        raise AxesBackendPermissionDenied(
            "AxesBackend detected that the given user is locked out"
        )


class AxesBackend(AxesStandaloneBackend, ModelBackend):
    """
    Axes authentication backend that also inherits from ModelBackend,
    and thus also performs other functions of ModelBackend such as permissions checks.

    Use this class as the first item of ``AUTHENTICATION_BACKENDS`` to
    prevent locked out users from being logged in by the Django authentication flow.

    .. note:: This backend does not log your user in. It monitors login attempts.
              Authentication is handled by the following backends that are configured in ``AUTHENTICATION_BACKENDS``.
    """



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("axes", "0003_auto_20160322_0929")]

    operations = [
        migrations.AlterModelOptions(
            name="accessattempt",
            options={
                "verbose_name": "access attempt",
                "verbose_name_plural": "access attempts",
            },
        ),
        migrations.AlterModelOptions(
            name="accesslog",
            options={
                "verbose_name": "access log",
                "verbose_name_plural": "access logs",
            },
        ),
        migrations.AlterField(
            model_name="accessattempt",
            name="attempt_time",
            field=models.DateTimeField(auto_now_add=True, verbose_name="Attempt Time"),
        ),
        migrations.AlterField(
            model_name="accessattempt",
            name="user_agent",
            field=models.CharField(
                db_index=True, max_length=255, verbose_name="User Agent"
            ),
        ),
        migrations.AlterField(
            model_name="accessattempt",
            name="username",
            field=models.CharField(
                db_index=True, max_length=255, null=True, verbose_name="Username"
            ),
        ),
        migrations.AlterField(
            model_name="accesslog",
            name="attempt_time",
            field=models.DateTimeField(auto_now_add=True, verbose_name="Attempt Time"),
        ),
        migrations.AlterField(
            model_name="accesslog",
            name="logout_time",
            field=models.DateTimeField(
                blank=True, null=True, verbose_name="Logout Time"
            ),
        ),
        migrations.AlterField(
            model_name="accesslog",
            name="user_agent",
            field=models.CharField(
                db_index=True, max_length=255, verbose_name="User Agent"
            ),
        ),
        migrations.AlterField(
            model_name="accesslog",
            name="username",
            field=models.CharField(
                db_index=True, max_length=255, null=True, verbose_name="Username"
            ),
        ),
    ]



from django.core.checks import run_checks, Warning  # pylint: disable=redefined-builtin
from django.test import override_settings, modify_settings

from axes.backends import AxesStandaloneBackend
from axes.checks import Messages, Hints, Codes
from tests.base import AxesTestCase


class CacheCheckTestCase(AxesTestCase):
    @override_settings(
        AXES_HANDLER="axes.handlers.cache.AxesCacheHandler",
        CACHES={
            "default": {
                "BACKEND": "django.core.cache.backends.db.DatabaseCache",
                "LOCATION": "axes_cache",
            }
        },
    )
    def test_cache_check(self):
        warnings = run_checks()
        self.assertEqual(warnings, [])

    @override_settings(
        AXES_HANDLER="axes.handlers.cache.AxesCacheHandler",
        CACHES={
            "default": {"BACKEND": "django.core.cache.backends.locmem.LocMemCache"}
        },
    )
    def test_cache_check_warnings(self):
        warnings = run_checks()
        warning = Warning(
            msg=Messages.CACHE_INVALID, hint=Hints.CACHE_INVALID, id=Codes.CACHE_INVALID
        )

        self.assertEqual(warnings, [warning])

    @override_settings(
        AXES_HANDLER="axes.handlers.database.AxesDatabaseHandler",
        CACHES={
            "default": {"BACKEND": "django.core.cache.backends.locmem.LocMemCache"}
        },
    )
    def test_cache_check_does_not_produce_check_warnings_with_database_handler(self):
        warnings = run_checks()
        self.assertEqual(warnings, [])


class MiddlewareCheckTestCase(AxesTestCase):
    @modify_settings(MIDDLEWARE={"remove": ["axes.middleware.AxesMiddleware"]})
    def test_cache_check_warnings(self):
        warnings = run_checks()
        warning = Warning(
            msg=Messages.MIDDLEWARE_INVALID,
            hint=Hints.MIDDLEWARE_INVALID,
            id=Codes.MIDDLEWARE_INVALID,
        )

        self.assertEqual(warnings, [warning])


class AxesSpecializedBackend(AxesStandaloneBackend):
    pass


class BackendCheckTestCase(AxesTestCase):
    @modify_settings(
        AUTHENTICATION_BACKENDS={"remove": ["axes.backends.AxesStandaloneBackend"]}
    )
    def test_backend_missing(self):
        warnings = run_checks()
        warning = Warning(
            msg=Messages.BACKEND_INVALID,
            hint=Hints.BACKEND_INVALID,
            id=Codes.BACKEND_INVALID,
        )

        self.assertEqual(warnings, [warning])

    @override_settings(
        AUTHENTICATION_BACKENDS=["tests.test_checks.AxesSpecializedBackend"]
    )
    def test_specialized_backend(self):
        warnings = run_checks()
        self.assertEqual(warnings, [])

    @override_settings(
        AUTHENTICATION_BACKENDS=["tests.test_checks.AxesNotDefinedBackend"]
    )
    def test_import_error(self):
        with self.assertRaises(ImportError):
            run_checks()

    @override_settings(AUTHENTICATION_BACKENDS=["module.not_defined"])
    def test_module_not_found_error(self):
        with self.assertRaises(ModuleNotFoundError):
            run_checks()


class DeprecatedSettingsTestCase(AxesTestCase):
    def setUp(self):
        self.disable_success_access_log_warning = Warning(
            msg=Messages.SETTING_DEPRECATED.format(
                deprecated_setting="AXES_DISABLE_SUCCESS_ACCESS_LOG"
            ),
            hint=Hints.SETTING_DEPRECATED,
            id=Codes.SETTING_DEPRECATED,
        )

    @override_settings(AXES_DISABLE_SUCCESS_ACCESS_LOG=True)
    def test_deprecated_success_access_log_flag(self):
        warnings = run_checks()
        self.assertEqual(warnings, [self.disable_success_access_log_warning])


class ConfCheckTestCase(AxesTestCase):
    @override_settings(AXES_USERNAME_CALLABLE="module.not_defined")
    def test_invalid_import_path(self):
        warnings = run_checks()
        warning = Warning(
            msg=Messages.CALLABLE_INVALID.format(
                callable_setting="AXES_USERNAME_CALLABLE"
            ),
            hint=Hints.CALLABLE_INVALID,
            id=Codes.CALLABLE_INVALID,
        )
        self.assertEqual(warnings, [warning])

    @override_settings(AXES_COOLOFF_TIME=lambda: 1)
    def test_valid_callable(self):
        warnings = run_checks()
        self.assertEqual(warnings, [])



from django.contrib import admin
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _

from axes.conf import settings
from axes.models import AccessAttempt, AccessLog, AccessFailureLog


class AccessAttemptAdmin(admin.ModelAdmin):
    list_display = (
        "attempt_time",
        "ip_address",
        "user_agent",
        "username",
        "path_info",
        "failures_since_start",
    )

    list_filter = ["attempt_time", "path_info"]

    search_fields = ["ip_address", "username", "user_agent", "path_info"]

    date_hierarchy = "attempt_time"

    fieldsets = (
        (None, {"fields": ("username", "path_info", "failures_since_start")}),
        (_("Form Data"), {"fields": ("get_data", "post_data")}),
        (_("Meta Data"), {"fields": ("user_agent", "ip_address", "http_accept")}),
    )

    readonly_fields = [
        "user_agent",
        "ip_address",
        "username",
        "http_accept",
        "path_info",
        "attempt_time",
        "get_data",
        "post_data",
        "failures_since_start",
    ]

    def has_add_permission(self, request: HttpRequest) -> bool:
        return False


class AccessLogAdmin(admin.ModelAdmin):
    list_display = (
        "attempt_time",
        "logout_time",
        "ip_address",
        "username",
        "user_agent",
        "path_info",
    )

    list_filter = ["attempt_time", "logout_time", "path_info"]

    search_fields = ["ip_address", "user_agent", "username", "path_info"]

    date_hierarchy = "attempt_time"

    fieldsets = (
        (None, {"fields": ("username", "path_info")}),
        (_("Meta Data"), {"fields": ("user_agent", "ip_address", "http_accept")}),
    )

    readonly_fields = [
        "user_agent",
        "ip_address",
        "username",
        "http_accept",
        "path_info",
        "attempt_time",
        "logout_time",
    ]

    def has_add_permission(self, request: HttpRequest) -> bool:
        return False


class AccessFailureLogAdmin(admin.ModelAdmin):
    list_display = (
        "attempt_time",
        "ip_address",
        "username",
        "user_agent",
        "path_info",
        "locked_out",
    )

    list_filter = ["attempt_time", "locked_out", "path_info"]

    search_fields = ["ip_address", "user_agent", "username", "path_info"]

    date_hierarchy = "attempt_time"

    fieldsets = (
        (None, {"fields": ("username", "path_info")}),
        (_("Meta Data"), {"fields": ("user_agent", "ip_address", "http_accept")}),
    )

    readonly_fields = [
        "user_agent",
        "ip_address",
        "username",
        "http_accept",
        "path_info",
        "attempt_time",
        "locked_out",
    ]

    def has_add_permission(self, request: HttpRequest) -> bool:
        return False


if settings.AXES_ENABLE_ADMIN:
    admin.site.register(AccessAttempt, AccessAttemptAdmin)
    admin.site.register(AccessLog, AccessLogAdmin)
    admin.site.register(AccessFailureLog, AccessFailureLogAdmin)



from django.core.checks import (  # pylint: disable=redefined-builtin
    Tags,
    Warning,
    register,
)
from django.utils.module_loading import import_string

from axes.backends import AxesStandaloneBackend
from axes.conf import settings


class Messages:
    CACHE_INVALID = (
        "You are using the django-axes cache handler for login attempt tracking."
        " Your cache configuration is however invalid and will not work correctly with django-axes."
        " This can leave security holes in your login systems as attempts are not tracked correctly."
        " Reconfigure settings.AXES_CACHE and settings.CACHES per django-axes configuration documentation."
    )
    MIDDLEWARE_INVALID = (
        "You do not have 'axes.middleware.AxesMiddleware' in your settings.MIDDLEWARE."
    )
    BACKEND_INVALID = "You do not have 'axes.backends.AxesStandaloneBackend' or a subclass in your settings.AUTHENTICATION_BACKENDS."
    SETTING_DEPRECATED = "You have a deprecated setting {deprecated_setting} configured in your project settings"
    CALLABLE_INVALID = "{callable_setting} is not a valid callable."


class Hints:
    CACHE_INVALID = None
    MIDDLEWARE_INVALID = None
    BACKEND_INVALID = "AxesModelBackend was renamed to AxesStandaloneBackend in django-axes version 5.0."
    SETTING_DEPRECATED = None
    CALLABLE_INVALID = None


class Codes:
    CACHE_INVALID = "axes.W001"
    MIDDLEWARE_INVALID = "axes.W002"
    BACKEND_INVALID = "axes.W003"
    SETTING_DEPRECATED = "axes.W004"
    CALLABLE_INVALID = "axes.W005"


@register(Tags.security, Tags.caches, Tags.compatibility)
def axes_cache_check(app_configs, **kwargs):  # pylint: disable=unused-argument
    axes_handler = getattr(settings, "AXES_HANDLER", "")

    axes_cache_key = getattr(settings, "AXES_CACHE", "default")
    axes_cache_config = settings.CACHES.get(axes_cache_key, {})
    axes_cache_backend = axes_cache_config.get("BACKEND", "")

    axes_cache_backend_incompatible = [
        "django.core.cache.backends.dummy.DummyCache",
        "django.core.cache.backends.locmem.LocMemCache",
        "django.core.cache.backends.filebased.FileBasedCache",
    ]

    warnings = []

    if axes_handler == "axes.handlers.cache.AxesCacheHandler":
        if axes_cache_backend in axes_cache_backend_incompatible:
            warnings.append(
                Warning(
                    msg=Messages.CACHE_INVALID,
                    hint=Hints.CACHE_INVALID,
                    id=Codes.CACHE_INVALID,
                )
            )

    return warnings


@register(Tags.security, Tags.compatibility)
def axes_middleware_check(app_configs, **kwargs):  # pylint: disable=unused-argument
    warnings = []

    if "axes.middleware.AxesMiddleware" not in settings.MIDDLEWARE:
        warnings.append(
            Warning(
                msg=Messages.MIDDLEWARE_INVALID,
                hint=Hints.MIDDLEWARE_INVALID,
                id=Codes.MIDDLEWARE_INVALID,
            )
        )

    return warnings


@register(Tags.security, Tags.compatibility)
def axes_backend_check(app_configs, **kwargs):  # pylint: disable=unused-argument
    warnings = []

    found = False
    for name in settings.AUTHENTICATION_BACKENDS:
        try:
            backend = import_string(name)
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                "Can not find module path defined in settings.AUTHENTICATION_BACKENDS"
            ) from e
        except ImportError as e:
            raise ImportError(
                "Can not import backend class defined in settings.AUTHENTICATION_BACKENDS"
            ) from e

        if issubclass(backend, AxesStandaloneBackend):
            found = True
            break

    if not found:
        warnings.append(
            Warning(
                msg=Messages.BACKEND_INVALID,
                hint=Hints.BACKEND_INVALID,
                id=Codes.BACKEND_INVALID,
            )
        )

    return warnings


@register(Tags.compatibility)
def axes_deprecation_check(app_configs, **kwargs):  # pylint: disable=unused-argument
    warnings = []

    deprecated_settings = [
        "AXES_DISABLE_SUCCESS_ACCESS_LOG",
        "AXES_LOGGER",
        # AXES_PROXY_ and AXES_META_ parameters were updated to more explicit
        # AXES_IPWARE_PROXY_ and AXES_IPWARE_META_ prefixes in version 6.x
        "AXES_PROXY_ORDER",
        "AXES_PROXY_COUNT",
        "AXES_PROXY_TRUSTED_IPS",
        "AXES_META_PRECEDENCE_ORDER",
        # AXES_ONLY_USER_FAILURES, AXES_USE_USER_AGENT and
        # AXES_LOCK_OUT parameters were replaced with AXES_LOCKOUT_PARAMETERS
        # in version 6.x
        "AXES_ONLY_USER_FAILURES",
        "AXES_LOCK_OUT_BY_USER_OR_IP",
        "AXES_LOCK_OUT_BY_COMBINATION_USER_AND_IP",
        "AXES_USE_USER_AGENT",
    ]

    for deprecated_setting in deprecated_settings:
        try:
            getattr(settings, deprecated_setting)
            warnings.append(
                Warning(
                    msg=Messages.SETTING_DEPRECATED.format(
                        deprecated_setting=deprecated_setting
                    ),
                    hint=None,
                    id=Codes.SETTING_DEPRECATED,
                )
            )
        except AttributeError:
            pass

    return warnings


@register
def axes_conf_check(app_configs, **kwargs):  # pylint: disable=unused-argument
    warnings = []

    callable_settings = [
        "AXES_CLIENT_IP_CALLABLE",
        "AXES_CLIENT_STR_CALLABLE",
        "AXES_LOCKOUT_CALLABLE",
        "AXES_USERNAME_CALLABLE",
        "AXES_WHITELIST_CALLABLE",
        "AXES_COOLOFF_TIME",
        "AXES_LOCKOUT_PARAMETERS",
    ]

    for callable_setting in callable_settings:
        value = getattr(settings, callable_setting)
        if not is_valid_callable(value):
            warnings.append(
                Warning(
                    msg=Messages.CALLABLE_INVALID.format(
                        callable_setting=callable_setting
                    ),
                    hint=Hints.CALLABLE_INVALID,
                    id=Codes.CALLABLE_INVALID,
                )
            )

    return warnings


def is_valid_callable(value) -> bool:
    if value is None:
        return True

    if callable(value):
        return True

    if isinstance(value, str):
        try:
            import_string(value)
        except ImportError:
            return False

    return True



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("axes", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="accessattempt",
            name="ip_address",
            field=models.GenericIPAddressField(
                db_index=True, null=True, verbose_name="IP Address"
            ),
        ),
        migrations.AlterField(
            model_name="accessattempt",
            name="trusted",
            field=models.BooleanField(db_index=True, default=False),
        ),
        migrations.AlterField(
            model_name="accessattempt",
            name="user_agent",
            field=models.CharField(db_index=True, max_length=255),
        ),
        migrations.AlterField(
            model_name="accessattempt",
            name="username",
            field=models.CharField(db_index=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name="accesslog",
            name="ip_address",
            field=models.GenericIPAddressField(
                db_index=True, null=True, verbose_name="IP Address"
            ),
        ),
        migrations.AlterField(
            model_name="accesslog",
            name="trusted",
            field=models.BooleanField(db_index=True, default=False),
        ),
        migrations.AlterField(
            model_name="accesslog",
            name="user_agent",
            field=models.CharField(db_index=True, max_length=255),
        ),
        migrations.AlterField(
            model_name="accesslog",
            name="username",
            field=models.CharField(db_index=True, max_length=255, null=True),
        ),
    ]



from random import choice
from string import ascii_letters, digits
from time import sleep

from django.contrib.auth import get_user_model
from django.http import HttpRequest
from django.test import TestCase
from django.urls import reverse
from django.utils.timezone import now

from axes.conf import settings
from axes.helpers import (
    get_cache,
    get_client_http_accept,
    get_client_ip_address,
    get_client_path_info,
    get_client_user_agent,
    get_cool_off,
    get_credentials,
    get_failure_limit,
)
from axes.models import AccessAttempt, AccessLog, AccessFailureLog
from axes.utils import reset


def custom_failure_limit(request, credentials):
    return 3


class AxesTestCase(TestCase):
    """
    Test case using custom settings for testing.
    """

    VALID_USERNAME = "axes-valid-username"
    VALID_PASSWORD = "axes-valid-password"
    VALID_EMAIL = "axes-valid-email@example.com"
    VALID_USER_AGENT = "axes-user-agent"
    VALID_IP_ADDRESS = "127.0.0.1"

    INVALID_USERNAME = "axes-invalid-username"
    INVALID_PASSWORD = "axes-invalid-password"
    INVALID_EMAIL = "axes-invalid-email@example.com"

    LOCKED_MESSAGE = "Account locked: too many login attempts."
    LOGOUT_MESSAGE = "Logged out"
    LOGIN_FORM_KEY = '<input type="submit" value="Log in" />'

    STATUS_SUCCESS = 200
    ALLOWED = 302
    BLOCKED = 429

    def setUp(self):
        """
        Create a valid user for login.
        """

        self.username = self.VALID_USERNAME
        self.password = self.VALID_PASSWORD
        self.email = self.VALID_EMAIL

        self.ip_address = self.VALID_IP_ADDRESS
        self.user_agent = self.VALID_USER_AGENT
        self.path_info = reverse("admin:login")

        self.user = get_user_model().objects.create_superuser(
            username=self.username, password=self.password, email=self.email
        )

        self.request = HttpRequest()
        self.request.method = "POST"
        self.request.META["REMOTE_ADDR"] = self.ip_address
        self.request.META["HTTP_USER_AGENT"] = self.user_agent
        self.request.META["PATH_INFO"] = self.path_info

        self.request.axes_attempt_time = now()
        self.request.axes_ip_address = get_client_ip_address(self.request)
        self.request.axes_user_agent = get_client_user_agent(self.request)
        self.request.axes_path_info = get_client_path_info(self.request)
        self.request.axes_http_accept = get_client_http_accept(self.request)
        self.request.axes_failures_since_start = None
        self.request.axes_locked_out = False

        self.credentials = get_credentials(self.username)

    def tearDown(self):
        get_cache().clear()

    def get_kwargs_with_defaults(self, **kwargs):
        defaults = {
            "user_agent": self.user_agent,
            "ip_address": self.ip_address,
            "username": self.username,
        }

        defaults.update(kwargs)
        return defaults

    def create_attempt(self, **kwargs):
        kwargs = self.get_kwargs_with_defaults(**kwargs)
        kwargs.setdefault("failures_since_start", 1)
        return AccessAttempt.objects.create(**kwargs)

    def create_log(self, **kwargs):
        return AccessLog.objects.create(**self.get_kwargs_with_defaults(**kwargs))

    def create_failure_log(self, **kwargs):
        return AccessFailureLog.objects.create(**self.get_kwargs_with_defaults(**kwargs))

    def reset(self, ip=None, username=None):
        return reset(ip, username)

    def login(
        self,
        is_valid_username=False,
        is_valid_password=False,
        remote_addr=None,
        **kwargs
    ):
        """
        Login a user.

        A valid credential is used when is_valid_username is True,
        otherwise it will use a random string to make a failed login.
        """

        if is_valid_username:
            username = self.VALID_USERNAME
        else:
            username = "".join(choice(ascii_letters + digits) for _ in range(10))

        if is_valid_password:
            password = self.VALID_PASSWORD
        else:
            password = self.INVALID_PASSWORD

        post_data = {"username": username, "password": password, **kwargs}

        return self.client.post(
            reverse("admin:login"),
            post_data,
            REMOTE_ADDR=remote_addr or self.ip_address,
            HTTP_USER_AGENT=self.user_agent,
        )

    def logout(self):
        return self.client.post(
            reverse("admin:logout"),
            REMOTE_ADDR=self.ip_address,
            HTTP_USER_AGENT=self.user_agent,
        )

    def check_login(self):
        response = self.login(is_valid_username=True, is_valid_password=True)
        self.assertNotContains(
            response, self.LOGIN_FORM_KEY, status_code=self.ALLOWED, html=True
        )

    def almost_lockout(self):
        for _ in range(1, get_failure_limit(None, None)):
            response = self.login()
            self.assertContains(response, self.LOGIN_FORM_KEY, html=True)

    def lockout(self):
        self.almost_lockout()
        return self.login()

    def check_lockout(self):
        response = self.lockout()
        if settings.AXES_LOCK_OUT_AT_FAILURE == True:
            self.assertContains(response, self.LOCKED_MESSAGE, status_code=self.BLOCKED)
        else:
            self.assertNotContains(
                response, self.LOCKED_MESSAGE, status_code=self.STATUS_SUCCESS
            )

    def cool_off(self):
        sleep(get_cool_off().total_seconds())

    def check_logout(self):
        response = self.logout()
        self.assertContains(
            response, self.LOGOUT_MESSAGE, status_code=self.STATUS_SUCCESS
        )

    def check_handler(self):
        """
        Check a handler and its basic functionality with lockouts, cool offs, login, and logout.

        This is a check that is intended to successfully run for each and every new handler.
        """

        self.check_lockout()
        self.cool_off()
        self.check_login()
        self.check_logout()



from logging import getLogger
from typing import Optional

from django.db import router, transaction
from django.db.models import F, Q, Sum, Value
from django.db.models.functions import Concat
from django.utils import timezone

from axes.attempts import (
    clean_expired_user_attempts,
    get_user_attempts,
    reset_user_attempts,
)
from axes.conf import settings
from axes.handlers.base import AxesBaseHandler, AbstractAxesHandler
from axes.helpers import (
    get_client_session_hash,
    get_client_str,
    get_client_username,
    get_credentials,
    get_failure_limit,
    get_lockout_parameters,
    get_query_str,
)
from axes.models import AccessLog, AccessAttempt, AccessFailureLog
from axes.signals import user_locked_out

log = getLogger(__name__)


class AxesDatabaseHandler(AbstractAxesHandler, AxesBaseHandler):
    """
    Signal handler implementation that records user login attempts to database and locks users out if necessary.

    .. note:: The get_user_attempts function is called several time during the authentication and lockout
              process, caching its output can be dangerous.
    """

    def reset_attempts(
        self,
        *,
        ip_address: Optional[str] = None,
        username: Optional[str] = None,
        ip_or_username: bool = False,
    ) -> int:
        attempts = AccessAttempt.objects.all()

        if ip_or_username:
            attempts = attempts.filter(Q(ip_address=ip_address) | Q(username=username))
        else:
            if ip_address:
                attempts = attempts.filter(ip_address=ip_address)
            if username:
                attempts = attempts.filter(username=username)

        count, _ = attempts.delete()
        log.info("AXES: Reset %d access attempts from database.", count)

        return count

    def reset_logs(self, *, age_days: Optional[int] = None) -> int:
        if age_days is None:
            count, _ = AccessLog.objects.all().delete()
            log.info("AXES: Reset all %d access logs from database.", count)
        else:
            limit = timezone.now() - timezone.timedelta(days=age_days)
            count, _ = AccessLog.objects.filter(attempt_time__lte=limit).delete()
            log.info(
                "AXES: Reset %d access logs older than %d days from database.",
                count,
                age_days,
            )

        return count

    def reset_failure_logs(self, *, age_days: Optional[int] = None) -> int:
        if age_days is None:
            count, _ = AccessFailureLog.objects.all().delete()
            log.info("AXES: Reset all %d access failure logs from database.", count)
        else:
            limit = timezone.now() - timezone.timedelta(days=age_days)
            count, _ = AccessFailureLog.objects.filter(attempt_time__lte=limit).delete()
            log.info(
                "AXES: Reset %d access failure logs older than %d days from database.",
                count,
                age_days,
            )

        return count

    def remove_out_of_limit_failure_logs(
        self,
        *,
        username: str,
        limit: Optional[int] = settings.AXES_ACCESS_FAILURE_LOG_PER_USER_LIMIT,
    ) -> int:
        count = 0
        failures = AccessFailureLog.objects.filter(username=username)
        out_of_limit_failures_logs = failures.count() - limit
        if out_of_limit_failures_logs > 0:
            for failure in failures[:out_of_limit_failures_logs]:
                failure.delete()
                count += 1
        return count

    def get_failures(self, request, credentials: Optional[dict] = None) -> int:
        attempts_list = get_user_attempts(request, credentials)
        attempt_count = max(
            (
                attempts.aggregate(Sum("failures_since_start"))[
                    "failures_since_start__sum"
                ]
                or 0
            )
            for attempts in attempts_list
        )
        return attempt_count

    def user_login_failed(self, sender, credentials: dict, request=None, **kwargs):
        """When user login fails, save AccessFailureLog record in database,
        save AccessAttempt record in database, mark request with
        lockout attribute and emit lockout signal.

        """

        log.info("AXES: User login failed, running database handler for failure.")

        if request is None:
            log.error(
                "AXES: AxesDatabaseHandler.user_login_failed does not function without a request."
            )
            return

        # 1. database query: Clean up expired user attempts from the database before logging new attempts
        clean_expired_user_attempts(request.axes_attempt_time)

        username = get_client_username(request, credentials)
        client_str = get_client_str(
            username,
            request.axes_ip_address,
            request.axes_user_agent,
            request.axes_path_info,
            request,
        )

        # If axes denied access, don't record the failed attempt as that would reset the lockout time.
        if (
            not settings.AXES_RESET_COOL_OFF_ON_FAILURE_DURING_LOCKOUT
            and request.axes_locked_out
        ):
            request.axes_credentials = credentials
            user_locked_out.send(
                "axes",
                request=request,
                username=username,
                ip_address=request.axes_ip_address,
            )
            return

        # This replaces null byte chars that crash saving failures.
        get_data = get_query_str(request.GET).replace("\0", "0x00")
        post_data = get_query_str(request.POST).replace("\0", "0x00")

        if self.is_whitelisted(request, credentials):
            log.info("AXES: Login failed from whitelisted client %s.", client_str)
            return

        # 2. database query: Get or create access record with the new failure data
        lockout_parameters = get_lockout_parameters(request, credentials)
        if lockout_parameters == ["username"] and username is None:
            log.warning(
                "AXES: Username is None and username is the only one lockout parameter, new record will NOT be created."
            )
        else:
            with transaction.atomic(using=router.db_for_write(AccessAttempt)):
                (
                    attempt,
                    created,
                ) = AccessAttempt.objects.select_for_update().get_or_create(
                    username=username,
                    ip_address=request.axes_ip_address,
                    user_agent=request.axes_user_agent,
                    defaults={
                        "get_data": get_data,
                        "post_data": post_data,
                        "http_accept": request.axes_http_accept,
                        "path_info": request.axes_path_info,
                        "failures_since_start": 1,
                        "attempt_time": request.axes_attempt_time,
                    },
                )

                # Record failed attempt with all the relevant information.
                # Filtering based on username, IP address and user agent handled elsewhere,
                # and this handler just records the available information for further use.
                if created:
                    log.warning(
                        "AXES: New login failure by %s. Created new record in the database.",
                        client_str,
                    )

                # 3. database query if there were previous attempts in the database
                # Update failed attempt information but do not touch the username, IP address, or user agent fields,
                # because attackers can request the site with multiple different configurations
                # in order to bypass the defense mechanisms that are used by the site.
                else:
                    separator = "\n---------\n"

                    attempt.get_data = Concat("get_data", Value(separator + get_data))
                    attempt.post_data = Concat(
                        "post_data", Value(separator + post_data)
                    )
                    attempt.http_accept = request.axes_http_accept
                    attempt.path_info = request.axes_path_info
                    attempt.failures_since_start = F("failures_since_start") + 1
                    attempt.attempt_time = request.axes_attempt_time
                    attempt.save()

                    log.warning(
                        "AXES: Repeated login failure by %s. Updated existing record in the database.",
                        client_str,
                    )

        # 3. or 4. database query: Calculate the current maximum failure number from the existing attempts
        failures_since_start = self.get_failures(request, credentials)
        request.axes_failures_since_start = failures_since_start

        if (
            settings.AXES_LOCK_OUT_AT_FAILURE
            and failures_since_start >= get_failure_limit(request, credentials)
        ):
            log.warning(
                "AXES: Locking out %s after repeated login failures.", client_str
            )

            request.axes_locked_out = True
            request.axes_credentials = credentials
            user_locked_out.send(
                "axes",
                request=request,
                username=username,
                ip_address=request.axes_ip_address,
            )

        # 5. database entry: Log for ever the attempt in the AccessFailureLog
        if settings.AXES_ENABLE_ACCESS_FAILURE_LOG:
            with transaction.atomic(using=router.db_for_write(AccessFailureLog)):
                AccessFailureLog.objects.create(
                    username=username,
                    ip_address=request.axes_ip_address,
                    user_agent=request.axes_user_agent,
                    http_accept=request.axes_http_accept,
                    path_info=request.axes_path_info,
                    attempt_time=request.axes_attempt_time,
                    locked_out=request.axes_locked_out,
                )
                self.remove_out_of_limit_failure_logs(username=username)

    def user_logged_in(self, sender, request, user, **kwargs):
        """
        When user logs in, update the AccessLog related to the user.
        """

        # 1. database query: Clean up expired user attempts from the database
        clean_expired_user_attempts(request.axes_attempt_time)

        username = user.get_username()
        credentials = get_credentials(username)
        client_str = get_client_str(
            username,
            request.axes_ip_address,
            request.axes_user_agent,
            request.axes_path_info,
            request,
        )

        log.info("AXES: Successful login by %s.", client_str)

        if not settings.AXES_DISABLE_ACCESS_LOG:
            # 2. database query: Insert new access logs with login time
            AccessLog.objects.create(
                username=username,
                ip_address=request.axes_ip_address,
                user_agent=request.axes_user_agent,
                http_accept=request.axes_http_accept,
                path_info=request.axes_path_info,
                attempt_time=request.axes_attempt_time,
                # evaluate session hash here to ensure having the correct
                # value which is stored on the backend
                session_hash=get_client_session_hash(request),
            )

        if settings.AXES_RESET_ON_SUCCESS:
            # 3. database query: Reset failed attempts for the logging in user
            count = reset_user_attempts(request, credentials)
            log.info(
                "AXES: Deleted %d failed login attempts by %s from database.",
                count,
                client_str,
            )

    def user_logged_out(self, sender, request, user, **kwargs):
        """
        When user logs out, update the AccessLog related to the user.
        """

        # 1. database query: Clean up expired user attempts from the database
        clean_expired_user_attempts(request.axes_attempt_time)

        username = user.get_username() if user else None
        client_str = get_client_str(
            username,
            request.axes_ip_address,
            request.axes_user_agent,
            request.axes_path_info,
            request,
        )

        log.info("AXES: Successful logout by %s.", client_str)

        if username and not settings.AXES_DISABLE_ACCESS_LOG:
            # 2. database query: Update existing attempt logs with logout time
            AccessLog.objects.filter(
                username=username,
                logout_time__isnull=True,
                # update only access log for given session
                session_hash=get_client_session_hash(request),
            ).update(logout_time=request.axes_attempt_time)

    def post_save_access_attempt(self, instance, **kwargs):
        """
        Handles the ``axes.models.AccessAttempt`` object post save signal.

        When needed, all post_save actions for this backend should be located
        here.
        """

    def post_delete_access_attempt(self, instance, **kwargs):
        """
        Handles the ``axes.models.AccessAttempt`` object post delete signal.

        When needed, all post_delete actions for this backend should be located
        here.
        """



# Generated by Django 2.0.4 on 2019-03-13 08:55

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("axes", "0005_remove_accessattempt_trusted")]

    operations = [migrations.RemoveField(model_name="accesslog", name="trusted")]



from axes.handlers.base import AxesBaseHandler, AbstractAxesHandler
from typing import Optional


class AxesDummyHandler(AbstractAxesHandler, AxesBaseHandler):
    """
    Signal handler implementation that does nothing and can be used to disable signal processing.
    """

    def is_allowed(self, request, credentials: Optional[dict] = None) -> bool:
        return True

    def user_login_failed(self, sender, credentials: dict, request=None, **kwargs):
        pass

    def user_logged_in(self, sender, request, user, **kwargs):
        pass

    def user_logged_out(self, sender, request, user, **kwargs):
        pass

    def get_failures(self, request, credentials: Optional[dict] = None) -> int:
        return 0



from django.core.management.base import BaseCommand

from axes.utils import reset


class Command(BaseCommand):
    help = "Reset all access attempts and lockouts"

    def handle(self, *args, **options):
        count = reset()

        if count:
            self.stdout.write(f"{count} attempts removed.")
        else:
            self.stdout.write("No attempts found.")



from axes.handlers.base import AxesHandler
from typing import Optional


class AxesTestHandler(AxesHandler):
    """
    Signal handler implementation that does nothing, ideal for a test suite.
    """

    def reset_attempts(
        self,
        *,
        ip_address: Optional[str] = None,
        username: Optional[str] = None,
        ip_or_username: bool = False,
    ) -> int:
        return 0

    def reset_logs(self, *, age_days: Optional[int] = None) -> int:
        return 0

    def is_allowed(self, request, credentials: Optional[dict] = None) -> bool:
        return True

    def get_failures(self, request, credentials: Optional[dict] = None) -> int:
        return 0



from axes.models import AccessFailureLog
from tests.base import AxesTestCase
from axes.helpers import get_failure_limit
from django.test import override_settings

@override_settings(AXES_ENABLE_ACCESS_FAILURE_LOG=True)
class FailureLogTestCase(AxesTestCase):
    def test_failure_log(self):
        self.login(is_valid_username=True, is_valid_password=False)
        self.assertEqual(AccessFailureLog.objects.count(), 1)
        self.assertTrue(AccessFailureLog.objects.filter(username=self.VALID_USERNAME).exists())
        self.assertTrue(AccessFailureLog.objects.filter(ip_address=self.ip_address).exists())

    def test_failure_locked_out(self):
        self.check_lockout()
        self.assertEqual(AccessFailureLog.objects.filter(locked_out=True).count(), 1)



# pylint: disable=import-outside-toplevel, unused-import

from logging import getLogger

from django import apps

from axes import __version__

log = getLogger(__name__)


class AppConfig(apps.AppConfig):
    default_auto_field = "django.db.models.AutoField"
    name = "axes"
    initialized = False

    @classmethod
    def initialize(cls):
        """
        Initialize Axes logging and show version information.

        This method is re-entrant and can be called multiple times.
        It displays version information exactly once at application startup.
        """

        if cls.initialized:
            return
        cls.initialized = True

        # Only import settings, checks, and signals one time after Django has been initialized
        from axes.conf import settings
        from axes import checks, signals

        # Skip startup log messages if Axes is not set to verbose
        if settings.AXES_VERBOSE:
            if callable(settings.AXES_LOCKOUT_PARAMETERS) or isinstance(
                settings.AXES_LOCKOUT_PARAMETERS, str
            ):
                mode = "blocking by parameters that are calculated in a custom callable"

            else:
                mode = "blocking by " + " or ".join(
                    [
                        (
                            param
                            if isinstance(param, str)
                            else "combination of " + " and ".join(param)
                        )
                        for param in settings.AXES_LOCKOUT_PARAMETERS
                    ]
                )

            log.info(
                "AXES: BEGIN version %s, %s",
                __version__,
                mode,
            )

    def ready(self):
        self.initialize()
